@@ -1,15 +1,27 @@
 import os
-import platform
 import struct
 import abc
 from itertools import count
 from multiprocessing import get_context
 from pickle import dumps, loads
-from threading import BrokenBarrierError
+from typing import Optional
 
 import trio
+from outcome import Outcome, capture
+
+
+class BrokenWorkerError(RuntimeError):
+    """Raised when a worker process fails or dies unexpectedly.
+
+    This error is not typically encountered in normal use, and indicates a severe
+    failure of either trio-parallel or the code that was executing in the worker.
+
+    The last argument of the exception is the underlying
+    :class:`multiprocessing.Process` which may be inspected for e.g. exit codes.
+    """
+
+    pass
 
-from ._util import BrokenWorkerError
 
 # How long a process will idle waiting for new work before gives up and exits.
 # This should be longer than a thread timeout proportionately to startup time.
@@ -20,28 +32,23 @@ _proc_counter = count()
 
 class WorkerProcBase(abc.ABC):
     def __init__(self, mp_context=get_context("spawn")):
-        # It is almost possible to synchronize on the pipe alone but on Pypy
-        # the _send_pipe doesn't raise the correct error on death. Anyway,
-        # this Barrier strategy is more obvious to understand.
-        self._barrier = mp_context.Barrier(2)
-        child_recv_pipe, self._send_pipe = mp_context.Pipe(duplex=False)
-        self._recv_pipe, child_send_pipe = mp_context.Pipe(duplex=False)
+        self._child_recv_pipe, self._send_pipe = mp_context.Pipe(duplex=False)
+        self._recv_pipe, self._child_send_pipe = mp_context.Pipe(duplex=False)
         self._proc = mp_context.Process(
             target=self._work,
-            args=(self._barrier, child_recv_pipe, child_send_pipe),
+            args=(self._child_recv_pipe, self._child_send_pipe),
             name=f"trio-parallel worker process {next(_proc_counter)}",
             daemon=True,
         )
-        # keep our own state flag for quick checks
-        self._started = False
+        self._started = trio.Event()  # Allow waiting before startup
+        self._wait_lock = trio.Lock()  # Efficiently multiplex waits
         self._rehabilitate_pipes()
 
     @staticmethod
-    def _work(barrier, recv_pipe, send_pipe):  # pragma: no cover
+    def _work(recv_pipe, send_pipe):  # pragma: no cover
 
         import inspect
         import signal
-        import outcome
 
         def coroutine_checker(fn, args):
             ret = fn(*args)
@@ -59,80 +66,92 @@ class WorkerProcBase(abc.ABC):
         # between processes. (Trio will take charge via cancellation.)
         signal.signal(signal.SIGINT, signal.SIG_IGN)
 
-        while True:
-            try:
-                # Return value is party #, not whether awoken within timeout
-                barrier.wait(timeout=IDLE_TIMEOUT)
-            except BrokenBarrierError:
-                # Timeout waiting for job, so we can exit.
-                return
-            # We got a job, and we are "woken"
-            fn, args = loads(recv_pipe.recv_bytes())
-            # Do the CPU bound work
-            result = outcome.capture(coroutine_checker, fn, args)
-            # Send result and go back to idling
-            send_pipe.send_bytes(dumps(result, protocol=-1))
-
-            del fn
-            del args
-            del result
-
-    async def run_sync(self, sync_fn, *args):
-        # Neither this nor the child process should be waiting at this point
-        assert not self._barrier.n_waiting, "Must first wake_up() the worker"
         try:
-            await self._send(dumps((sync_fn, args), protocol=-1))
-            # noinspection PyTypeChecker
-            result = loads(await self._recv())
-        except trio.EndOfChannel:
-            # Likely the worker died while we were waiting on a pipe
-            self.kill()  # NOTE: must reap zombie child elsewhere
-            raise BrokenWorkerError(f"{self._proc} died unexpectedly")
-        except BaseException:
-            # Cancellation or other unknown errors leave the process in an
-            # unknown state, so there is no choice but to kill.
-            self.kill()  # NOTE: must reap zombie child elsewhere
-            raise
-        else:
-            return result.unwrap()
+            while recv_pipe.poll(IDLE_TIMEOUT):
+                fn, args = loads(recv_pipe.recv_bytes())
+                # Do the CPU bound work
+                result = capture(coroutine_checker, fn, args)
+                # Send result and go back to idling
+                send_pipe.send_bytes(dumps(result, protocol=-1))
+
+                del fn
+                del args
+                del result
+        except (BrokenPipeError, EOFError):
+            # If the main process ends but this one is still alive, we will
+            # observe one of these exceptions and can simply exit quietly.
+            return
+
+        # Clean idle shutdown: close recv_pipe first to minimize subsequent race.
+        recv_pipe.close()
+        # Race condition: it is possible to sneak a write through in the main process
+        # between the recv_pipe poll timeout and close. Naively, this would
+        # make a clean shutdown look like a broken worker. By sending a sentinel
+        # value, we can indicate to a waiting main process that we have hit this
+        # race condition and need a restart. However, the send MUST be non-blocking
+        # to free this process's resources in a timely manner. Therefore, this message
+        # can be any size on Windows but must be less than 512 bytes by POSIX.1-2001.
+        send_pipe.send_bytes(dumps(None, protocol=-1))
+        send_pipe.close()
+
+    async def run_sync(self, sync_fn, *args) -> Optional[Outcome]:
+        result = None
+        try:
+
+            if not self._started.is_set():
+                await trio.to_thread.run_sync(self._proc.start)
+                # XXX: We must explicitly close these after start to see child closures
+                self._child_send_pipe.close()
+                self._child_recv_pipe.close()
+                self._started.set()
+
+            try:
+                await self._send(dumps((sync_fn, args), protocol=-1))
+            except trio.BrokenResourceError:
+                return None
+
+            try:
+                result = loads(await self._recv())
+            except trio.EndOfChannel:
+                result = await self.wait()  # skip kill/wait in finally block
+                raise BrokenWorkerError(
+                    "Worker died unexpectedly:", self._proc
+                ) from None
+
+            return result
+
+        finally:  # pragma: no cover convoluted branching, but the concept is simple
+            if result is None:  # that is, if anything interrupted a normal result
+                self.kill()  # maybe redundant
+                # do cleanup soon, but no need to block here
+                trio.lowlevel.spawn_system_task(self.wait)
 
     def is_alive(self):
         # if the proc is alive, there is a race condition where it could be
-        # dying, but the the barrier should be broken at that time. This
-        # call reaps zombie children on Unix.
+        # dying. This call reaps zombie children on Unix.
         return self._proc.is_alive()
 
-    def wake_up(self, timeout=None):
-        if not self._started:
-            self._proc.start()
-            self._started = True
-        try:
-            self._barrier.wait(timeout)
-        except BrokenBarrierError:
-            # raise our own flavor of exception and reap child
-            if self._proc.is_alive():  # pragma: no cover - rare race condition
-                self.kill()
-                self._proc.join(1)  # this will block for ms, but it should be rare
-                if self._proc.is_alive():
-                    raise RuntimeError(f"{self._proc} alive after failed wakeup")
-            raise BrokenWorkerError(f"{self._proc} died unexpectedly") from None
-
     def kill(self):
-        # race condition: if we kill while the proc has the underlying
-        # semaphore, we can deadlock it, so make sure we hold it.
-        with self._barrier._cond:
-            self._barrier.abort()
-            try:
-                self._proc.kill()
-            except AttributeError:
-                # cpython 3.6 has an edge case where if this process holds
-                # the semaphore, a wait can timeout and raise an OSError.
-                self._proc.terminate()
+        if self._proc.pid is None:
+            self._started.set()  # unblock self.wait()
+            return
+        try:
+            self._proc.kill()
+        except AttributeError:
+            self._proc.terminate()
 
-    @abc.abstractmethod
     async def wait(self):
-        pass
+        await self._started.wait()
+        if self._proc.pid is None:
+            return None  # killed before started
+        async with self._wait_lock:
+            await self._wait()
+        # redundant wait, but also fix a macos race and do Process cleanup
+        self._proc.join()
+        # unfortunately join does not return exitcode
+        return self._proc.exitcode
 
+    # platform-specific behavior
     @abc.abstractmethod
     def _rehabilitate_pipes(self):
         pass
@@ -145,11 +164,14 @@ class WorkerProcBase(abc.ABC):
     async def _send(self, buf):
         pass
 
+    @abc.abstractmethod
+    async def _wait(self):
+        pass
+
 
 class WindowsWorkerProc(WorkerProcBase):
-    async def wait(self):
+    async def _wait(self):
         await trio.lowlevel.WaitForSingleObject(self._proc.sentinel)
-        return self._proc.exitcode
 
     def _rehabilitate_pipes(self):
         # These must be created in an async context
@@ -175,16 +197,8 @@ class WindowsWorkerProc(WorkerProcBase):
 
 
 class PosixWorkerProc(WorkerProcBase):
-    async def wait(self):
+    async def _wait(self):
         await trio.lowlevel.wait_readable(self._proc.sentinel)
-        e = self._proc.exitcode
-        if e is not None:
-            return e
-        else:  # pragma: no cover # to avoid flaky CI, but it happens regularly
-            # race on macOS, see comment in trio.Process._wait
-            self._proc.join()
-            # unfortunately join does not return exitcode
-            return self._proc.exitcode
 
     def _rehabilitate_pipes(self):
         # These must be created in an async context
@@ -246,34 +260,12 @@ class PosixWorkerProc(WorkerProcBase):
             pass
 
 
-class PypyWorkerProc(WorkerProcBase):
-    async def run_sync(self, sync_fn, *args):
-        async with trio.open_nursery() as nursery:
-            nursery.start_soon(self._child_monitor)
-            result = await super().run_sync(sync_fn, *args)
-            nursery.cancel_scope.cancel()
-        return result
-
-    async def _child_monitor(self):
-        # If this worker dies, raise a catchable error...
-        await self.wait()
-        # but not if another error or cancel is incoming, those take priority!
-        await trio.lowlevel.checkpoint_if_cancelled()
-        raise BrokenWorkerError(f"{self._proc} died unexpectedly")
-
-
 if os.name == "nt":
 
     class WorkerProc(WindowsWorkerProc):
         pass
 
 
-elif platform.python_implementation() == "PyPy":
-
-    class WorkerProc(PypyWorkerProc, PosixWorkerProc):
-        pass
-
-
 else:
 
     class WorkerProc(PosixWorkerProc):
